import ast
import importlib
import inspect
import os
import re
import shutil
from collections.abc import Callable, Collection, Generator, Sequence
from copy import deepcopy
from dataclasses import dataclass
from datetime import date
from enum import Enum, auto
from pathlib import Path
from types import GenericAlias, LambdaType, ModuleType, NoneType
from typing import (
    Any,
    TypeAlias,
    _BaseGenericAlias,  # pyright: ignore[reportGeneralTypeIssues]
    final,
    get_args,
    get_origin,
)

from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo, ModelField, ModelPrivateAttr
from typing_extensions import assert_never

from cadwyn.structure.enums import (
    AlterEnumSubInstruction,
    EnumDidntHaveMembersInstruction,
    EnumHadMembersInstruction,
)
from cadwyn.structure.schemas import (
    AlterSchemaInstruction,
    AlterSchemaSubInstruction,
    OldSchemaFieldDidntExist,
    OldSchemaFieldExistedWith,
    OldSchemaFieldHad,
)
from cadwyn.structure.versions import Version, VersionBundle

from ._utils import Sentinel, UnionType, get_index_of_base_schema_dir_in_pythonpath
from .exceptions import CodeGenerationError, InvalidGenerationInstructionError

_LambdaFunctionName = (lambda: None).__name__  # pragma: no branch
_FieldName: TypeAlias = str
_empty_field_info = Field()
_dict_of_empty_field_info = {k: getattr(_empty_field_info, k) for k in FieldInfo.__slots__}
_RE_CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")
_AUTO_GENERATION_WARNING = "# THIS FILE WAS AUTO-GENERATED BY CADWYN. DO NOT EVER TRY TO EDIT IT BY HAND\n\n"


@final
class _ImportedModule:
    __slots__ = (
        "path",
        "name",
        "alias",
        "absolute_python_path_to_origin",
        "how_far_up_is_base_schema_dir_from_current_module",
        "is_package",
    )

    def __init__(
        self,
        version_dir: str,
        import_pythonpath_template: str,
        package_name: str,
        how_far_up_is_base_schema_dir_from_current_module: int,
        absolute_python_path_template: str,
        is_package: bool,
    ) -> None:
        self.path = import_pythonpath_template.format(version_dir)
        self.name = package_name.format(version_dir)
        if self.path == "":
            self.alias = self.name
        else:
            self.alias = f"{self.path.replace('.', '_')}_{self.name}"
        self.absolute_python_path_to_origin = absolute_python_path_template.format("latest")
        self.how_far_up_is_base_schema_dir_from_current_module = how_far_up_is_base_schema_dir_from_current_module
        self.is_package = is_package

    def get_ast(self) -> ast.ImportFrom:
        module = f"{self.path}.{self.name}"
        name = ast.alias(name="*")
        level = self.how_far_up_is_base_schema_dir_from_current_module
        # TODO: Add a testcase where is_package == True and level == 3
        if self.is_package and level == 2:
            level -= 1
        return ast.ImportFrom(
            level=level,
            module=module,
            names=[name],
        )


@dataclass(slots=True)
class ModelFieldLike:
    name: str
    original_type: Any
    annotation: Any
    field_info: FieldInfo
    import_from: str | None
    import_as: str | None


@dataclass(slots=True)
class ModelPrivateAttrInfo:
    attr: ModelPrivateAttr
    annotation: Any


@dataclass(slots=True)
class ModelInfo:
    name: str
    fields: dict[_FieldName, tuple[type[BaseModel], ModelField | ModelFieldLike]]


@dataclass(slots=True)
class _SchemaBundle:
    version: date
    schemas: dict[str, ModelInfo]


def generate_code_for_versioned_packages(
    template_module: ModuleType,
    versions: VersionBundle,
    *,
    ignore_coverage_for_latest_aliases: bool = True,
):
    """_summary_

    Args:
        template_module: The latest package from which we will generate the versioned packages
        versions: Version bundle to generate versions from
        ignore_coverage_for_latest_aliases: Add a pragma: no cover comment to the star imports in the generated
        version of the latest module.
    """
    schemas = {
        k: ModelInfo(v.__name__, _get_fields_for_model(v)) for k, v in deepcopy(versions.versioned_schemas).items()
    }
    enums = {k: (v, {member.name: member.value for member in v}) for k, v in deepcopy(versions.versioned_enums).items()}
    schemas_per_version: list[_SchemaBundle] = []
    version_list = list(versions)

    # The first one is for the "latest"
    schemas_per_version: list[_SchemaBundle] = []
    version_list = list(versions)

    schemas_per_version.append(_SchemaBundle(version_list[0].value, schemas))
    # Special casing for the first first because it is equivalent to latest
    _generate_latest_version_alias_directory(
        version_list[0].value,
        template_module,
        ignore_coverage_for_latest_aliases,
    )
    schemas = deepcopy(schemas)
    _apply_migrations(version_list[0], schemas, enums)
    for version in version_list[1:]:
        schemas_per_version.append(_SchemaBundle(version.value, schemas))
        _generate_versioned_directory(template_module, schemas, enums, version.value)
        schemas = deepcopy(schemas)
        _apply_migrations(version, schemas, enums)


def _generate_latest_version_alias_directory(
    version: date,
    template_module: ModuleType,
    ignore_coverage_for_latest_aliases: bool,
):
    version_dir_name = _get_version_dir_name(version)
    template_dir = _get_package_path_from_module(template_module)
    version_dir = template_dir.with_name(version_dir_name)
    index_of_latest_schema_dir_in_pythonpath = get_index_of_base_schema_dir_in_pythonpath(
        template_module,
        version_dir,
    )
    for _, original_module, parallel_file in _generate_parallel_directory(template_module, version_dir):
        imports = _prepare_imports_from_version_dirs(
            original_module,
            ["latest"],
            index_of_latest_schema_dir_in_pythonpath,
        )

        import_text = _AUTO_GENERATION_WARNING + ast.unparse(imports[0].get_ast()) + " # noqa: F403"
        if ignore_coverage_for_latest_aliases:
            import_text += " # pragma: no cover"

        parallel_file.write_text(import_text)


def _prepare_imports_from_version_dirs(
    original_module: ModuleType,
    version_dir_names: Collection[str],
    index_of_latest_schema_dir_in_pythonpath: int,
) -> list[_ImportedModule]:
    # package.latest                     -> from .. import latest
    # package.latest.module              -> from ...latest import module
    # package.latest.subpackage          -> from ...latest import subpackage
    # package.latest.subpackage.module   -> from ....subpackage import module

    # package.latest                    -> from ..latest import *
    # package.latest.module             -> from ..latest.module import *
    # package.latest.subpackage         -> from ...latest.subpackage import *
    # package.latest.subpackage.module  -> from ...latest.subpackage.module import *

    original_module_parts = original_module.__name__.split(".")
    original_module_parts[index_of_latest_schema_dir_in_pythonpath] = "{}"
    how_far_up_is_base_schema_dir_from_current_module = (
        len(original_module_parts) - index_of_latest_schema_dir_in_pythonpath
    )
    is_package = original_module_parts[-1] == "__init__"
    if is_package:
        original_module_parts.pop(-1)

    package_name = original_module_parts[-1]
    package_path = original_module_parts[index_of_latest_schema_dir_in_pythonpath:-1]
    import_pythonpath_template = ".".join(package_path)
    absolute_python_path_template = ".".join(original_module_parts)
    return [
        _ImportedModule(
            version_dir,
            import_pythonpath_template,
            package_name,
            how_far_up_is_base_schema_dir_from_current_module,
            absolute_python_path_template,
            is_package,
        )
        for version_dir in version_dir_names
    ]


def _apply_migrations(
    version: Version,
    schemas: dict[
        str,
        ModelInfo,
    ],
    enums: dict[str, tuple[type[Enum], dict[str, Any]]],
):
    for version_change in version.version_changes:
        _apply_alter_schema_instructions(
            schemas,
            version_change.alter_schema_instructions,
            version_change.__name__,
        )
        _apply_alter_enum_instructions(
            enums,
            version_change.alter_enum_instructions,
            version_change.__name__,
        )


def _apply_alter_schema_instructions(  # noqa: C901
    schema_infos: dict[str, ModelInfo],
    alter_schema_instructions: Sequence[AlterSchemaSubInstruction | AlterSchemaInstruction],
    version_change_name: str,
):
    # TODO: If we have a request migration for an endpoint instead of a schema and we haven't found that endpoint
    # during codegen -- raise an error or maybe add an argument that controlls that. Or maybe this is overengineering..
    for alter_schema_instruction in alter_schema_instructions:
        schema = alter_schema_instruction.schema
        schema_path = _get_cls_pythonpath(schema)
        mutable_schema_info = schema_infos[schema_path]
        field_name_to_field_model = mutable_schema_info.fields
        if isinstance(alter_schema_instruction, OldSchemaFieldDidntExist):
            if alter_schema_instruction.field_name not in field_name_to_field_model:
                raise InvalidGenerationInstructionError(
                    f'You tried to delete a field "{alter_schema_instruction.field_name}" from "{schema.__name__}" '
                    f'in "{version_change_name}" but it doesn\'t have such a field.',
                )
            field_name_to_field_model.pop(alter_schema_instruction.field_name)

        elif isinstance(alter_schema_instruction, OldSchemaFieldHad):
            if alter_schema_instruction.field_name not in field_name_to_field_model:
                raise InvalidGenerationInstructionError(
                    f'You tried to change the type of field "{alter_schema_instruction.field_name}" from '
                    f'"{schema.__name__}" in "{version_change_name}" but it doesn\'t have such a field.',
                )
            model_field = field_name_to_field_model[alter_schema_instruction.field_name][1]
            if alter_schema_instruction.type is not Sentinel:
                if model_field.annotation == alter_schema_instruction.type:
                    raise InvalidGenerationInstructionError(
                        f'You tried to change the type of field "{alter_schema_instruction.field_name}" to '
                        f'"{alter_schema_instruction.type}" from "{schema.__name__}" in "{version_change_name}" '
                        f'but it already has type "{model_field.annotation}"',
                    )
                model_field.annotation = alter_schema_instruction.type
            if alter_schema_instruction.new_name is not Sentinel:
                if alter_schema_instruction.new_name == alter_schema_instruction.field_name:
                    raise InvalidGenerationInstructionError(
                        f'You tried to change the name of field "{alter_schema_instruction.field_name}" '
                        f'from "{schema.__name__}" in "{version_change_name}" '
                        "but it already has that name.",
                    )
                field_name_to_field_model[alter_schema_instruction.new_name] = field_name_to_field_model.pop(
                    alter_schema_instruction.field_name,
                )
            field_info = model_field.field_info

            dict_of_field_info = {k: getattr(field_info, k) for k in field_info.__slots__}
            if dict_of_field_info == _dict_of_empty_field_info:
                field_info = FieldInfo()
                model_field.field_info = field_info
            for attr_name in alter_schema_instruction.field_changes.__dataclass_fields__:
                attr_value = getattr(alter_schema_instruction.field_changes, attr_name)
                if attr_value is not Sentinel:
                    if getattr(field_info, attr_name) == attr_value:
                        raise InvalidGenerationInstructionError(
                            f'You tried to change the attribute "{attr_name}" of field '
                            f'"{alter_schema_instruction.field_name}" '
                            f'from "{schema.__name__}" to {attr_value!r} in "{version_change_name}" '
                            "but it already has that value.",
                        )
                    setattr(field_info, attr_name, attr_value)
        elif isinstance(alter_schema_instruction, OldSchemaFieldExistedWith):
            if alter_schema_instruction.field_name in field_name_to_field_model:
                raise InvalidGenerationInstructionError(
                    f'You tried to add a field "{alter_schema_instruction.field_name}" to "{schema.__name__}" '
                    f'in "{version_change_name}" but there is already a field with that name.',
                )
            if alter_schema_instruction.import_as is not None:
                annotation = alter_schema_instruction.import_as
            else:
                annotation = alter_schema_instruction.type
            field_name_to_field_model[alter_schema_instruction.field_name] = (
                schema,
                ModelFieldLike(
                    name=alter_schema_instruction.field_name,
                    original_type=alter_schema_instruction.type,
                    annotation=annotation,
                    field_info=alter_schema_instruction.field,
                    import_from=alter_schema_instruction.import_from,
                    import_as=alter_schema_instruction.import_as,
                ),
            )
        elif isinstance(alter_schema_instruction, AlterSchemaInstruction):
            # We only handle names right now so we just go ahead and check
            if alter_schema_instruction.name == mutable_schema_info.name:
                raise InvalidGenerationInstructionError(
                    f'You tried to change the name of "{schema.__name__}" in "{version_change_name}" '
                    "but it already has the name you tried to assign.",
                )
            mutable_schema_info.name = alter_schema_instruction.name
        else:
            assert_never(alter_schema_instruction)


def _apply_alter_enum_instructions(
    enums: dict[str, tuple[type[Enum], dict[str, Any]]],
    alter_enum_instructions: Sequence[AlterEnumSubInstruction],
    version_change_name: str,
):
    for alter_enum_instruction in alter_enum_instructions:
        enum = alter_enum_instruction.enum
        enum_path = _get_cls_pythonpath(enum)
        enum_member_to_value = enums[enum_path]
        if isinstance(alter_enum_instruction, EnumDidntHaveMembersInstruction):
            for member in alter_enum_instruction.members:
                if member not in enum_member_to_value[1]:
                    raise InvalidGenerationInstructionError(
                        f'You tried to delete a member "{member}" from "{enum.__name__}" '
                        f'in "{version_change_name}" but it doesn\'t have such a member.',
                    )
                enum_member_to_value[1].pop(member)
        elif isinstance(alter_enum_instruction, EnumHadMembersInstruction):
            for member, member_value in alter_enum_instruction.members.items():
                if member in enum_member_to_value[1] and enum_member_to_value[1][member] == member_value:
                    raise InvalidGenerationInstructionError(
                        f'You tried to add a member "{member}" to "{enum.__name__}" '
                        f'in "{version_change_name}" but there is already a member with that name and value.',
                    )
                enum_member_to_value[1][member] = member_value
        else:
            assert_never(alter_enum_instruction)


def _get_version_dir_path(template_module: ModuleType, version: date) -> Path:
    template_dir = _get_package_path_from_module(template_module)
    return template_dir.with_name(_get_version_dir_name(version))


def _get_version_dir_name(version: date):
    return "v" + version.isoformat().replace("-", "_")


# TODO OPTIMIZATION: This is called way too many times. Make it so we call it only once
def _get_package_path_from_module(template_module: ModuleType) -> Path:
    file = inspect.getsourcefile(template_module)

    # I am too lazy to reproduce this error correctly
    if file is None:  # pragma: no cover
        raise CodeGenerationError(f'Module "{template_module}" has no source file')
    file = Path(file)
    if not file.name == "__init__.py":
        raise CodeGenerationError(f'Module "{template_module}" is not a package')
    return file.parent


def _generate_versioned_directory(
    template_module: ModuleType,
    schemas: dict[str, ModelInfo],
    enums: dict[str, tuple[type[Enum], dict[str, Any]]],
    version: date,
):
    version_dir = _get_version_dir_path(template_module, version)
    for (
        _relative_path_to_file,
        original_module,
        parallel_file,
    ) in _generate_parallel_directory(
        template_module,
        version_dir,
    ):
        new_module_text = _migrate_module_to_another_version(
            original_module,
            schemas,
            enums,
        )
        parallel_file.write_text(_AUTO_GENERATION_WARNING + new_module_text)


def _generate_parallel_directory(
    template_module: ModuleType,
    parallel_dir: Path,
) -> Generator[tuple[Path, ModuleType, Path], Any, None]:
    if template_module.__file__ is None:  # pragma: no cover
        raise ValueError(
            f"You passed a {template_module=} but it doesn't have a file "
            "so it is impossible to generate its counterpart.",
        )
    dir = _get_package_path_from_module(template_module)
    parallel_dir.mkdir(exist_ok=True)
    # >>> [cadwyn, structure, schemas]
    template_module_python_path_parts = template_module.__name__.split(".")
    # >>> [home, foo, bar, cadwyn, structure, schemas]
    template_module_path_parts = Path(template_module.__file__).parent.parts
    # >>> [home, foo, bar] = [home, foo, bar, cadwyn, structure, schemas][:-3]
    root_module_path = Path(
        *template_module_path_parts[: -len(template_module_python_path_parts)],
    )
    for subroot, dirnames, filenames in os.walk(dir):
        original_subroot = Path(subroot)
        parallel_subroot = parallel_dir / original_subroot.relative_to(dir)
        if "__pycache__" in dirnames:
            dirnames.remove("__pycache__")
        for dirname in dirnames:
            (parallel_subroot / dirname).mkdir(exist_ok=True)
        for filename in filenames:
            original_file = (original_subroot / filename).absolute()
            parallel_file = (parallel_subroot / filename).absolute()

            if filename.endswith(".py"):
                original_module_path = ".".join(
                    original_file.relative_to(root_module_path).with_suffix("").parts,
                )
                original_module = importlib.import_module(original_module_path)
                yield original_subroot.relative_to(dir), original_module, parallel_file
            else:
                shutil.copyfile(original_file, parallel_file)


def _get_fields_for_model(
    model: type[BaseModel],
) -> dict[_FieldName, tuple[type[BaseModel], ModelField | ModelFieldLike]]:
    actual_fields: dict[_FieldName, tuple[type[BaseModel], ModelField | ModelFieldLike]] = {}
    for cls in model.__mro__:
        if cls is BaseModel:
            return actual_fields
        if not issubclass(cls, BaseModel):
            continue
        for field_name, field in cls.__fields__.items():
            if field_name not in actual_fields and field_name in cls.__annotations__:
                actual_fields[field_name] = (cls, field)
    raise CodeGenerationError(f"Model {model} is not a subclass of BaseModel")


def _parse_python_module(module: ModuleType) -> ast.Module:
    try:
        return ast.parse(inspect.getsource(module))
    except OSError as e:
        if module.__file__ is None:  # pragma: no cover
            raise CodeGenerationError("Failed to get file path to the module") from e

        path = Path(module.__file__)
        if path.is_file() and path.read_text() == "":
            return ast.Module([])
        # Not sure how to get here so this is just a precaution
        raise CodeGenerationError(
            "Failed to get source code for module",
        ) from e  # pragma: no cover


def _migrate_module_to_another_version(
    module: ModuleType,
    modified_schemas: dict[str, ModelInfo],
    modified_enums: dict[str, tuple[type[Enum], dict[str, Any]]],
) -> str:
    transformer = _AnnotationTransformer()
    parsed_file = _parse_python_module(module)
    if module.__name__.endswith(".__init__"):
        module_name = module.__name__.removesuffix(".__init__")
    else:
        module_name = module.__name__
    all_names_in_file = _get_all_names_defined_in_module(parsed_file, module_name)

    # TODO: Does this play well with renaming?
    extra_field_imports = [
        ast.ImportFrom(
            module=field.import_from,
            names=[ast.alias(name=transformer.visit(field.original_type).strip("'"), asname=field.import_as)],
            level=0,
        )
        for val in modified_schemas.values()
        for _, field in val.fields.values()
        if isinstance(field, ModelFieldLike) and field.import_from is not None
    ]

    body = ast.Module(
        [
            ast.ImportFrom(module="pydantic", names=[ast.alias(name="Field")], level=0),
            ast.Import(names=[ast.alias(name="typing")], level=0),
            ast.ImportFrom(module="typing", names=[ast.alias(name="Any")], level=0),
        ]
        + extra_field_imports
        + [
            _migrate_ast_node_to_another_version(
                all_names_in_file,
                n,
                module_name,
                modified_schemas,
                modified_enums,
            )
            for n in parsed_file.body
        ],
        [],
    )

    return ast.unparse(body)


def _migrate_ast_node_to_another_version(
    all_names_in_module: dict[str, str],
    node: ast.stmt,
    module_python_path: str,
    modified_schemas: dict[str, ModelInfo],
    modified_enums: dict[str, tuple[type[Enum], dict[str, Any]]],
):
    if isinstance(node, ast.ClassDef):
        return _migrate_cls_to_another_version(
            all_names_in_module,
            node,
            module_python_path,
            modified_schemas,
            modified_enums,
        )
    elif isinstance(node, ast.ImportFrom):
        python_path = _get_absolute_python_path_of_import(node, module_python_path)
        node.names = [
            name
            if (name_path := f"{python_path}.{name.name}") not in modified_schemas
            else ast.alias(name=modified_schemas[name_path].name, asname=name.asname)
            for name in node.names
        ]

    return node


def _get_absolute_python_path_of_import(node: ast.ImportFrom, module_python_path: str):
    python_path = ".".join(module_python_path.split(".")[0 : -node.level])
    result = []
    if node.module:
        result.append(node.module)
    if python_path:
        result.append(python_path)
    return ".".join(result)


def _migrate_cls_to_another_version(
    all_names_in_module: dict[str, str],
    cls_node: ast.ClassDef,
    module_python_path: str,
    modified_schemas: dict[str, ModelInfo],
    modified_enums: dict[str, tuple[type[Enum], dict[str, Any]]],
) -> ast.ClassDef:
    cls_python_path = f"{module_python_path}.{cls_node.name}"
    cls_node = _modify_schema_cls(
        all_names_in_module,
        cls_node,
        modified_schemas,
        module_python_path,
        cls_python_path,
    )
    if cls_python_path in modified_enums:
        cls_node = _modify_enum_cls(cls_node, modified_enums[cls_python_path][1])

    if not cls_node.body:
        cls_node.body = [ast.Pass()]
    return cls_node


# TODO: Make sure that cadwyn doesn't remove OLD property definitions
def _modify_schema_cls(
    all_names_in_module: dict[str, str],
    cls_node: ast.ClassDef,
    modified_schemas: dict[str, ModelInfo],
    module_python_path: str,
    cls_python_path: str,
) -> ast.ClassDef:
    annotation_transformer = _AnnotationTransformerWithSchemaRenaming(
        modified_schemas,
        module_python_path,
        all_names_in_module,
    )
    ast_transformer = _AnnotationASTNodeTransformer(modified_schemas, all_names_in_module, module_python_path)
    if cls_python_path in modified_schemas:
        model_info = modified_schemas[cls_python_path]
        cls_node.name = model_info.name
        field_definitions = [
            ast.AnnAssign(
                target=ast.Name(name, ctx=ast.Store()),
                annotation=ast.Name(annotation_transformer.visit(field.annotation)),
                value=_generate_field_ast(field, annotation_transformer),
                simple=1,
            )
            for name, (_, field) in model_info.fields.items()
        ]
    else:
        field_definitions = [field for field in cls_node.body if isinstance(field, ast.AnnAssign)]

    old_body = [n for n in cls_node.body if not isinstance(n, ast.AnnAssign | ast.Pass | ast.Ellipsis)]
    docstring = _pop_docstring_from_cls_body(old_body)
    cls_node.body = docstring + field_definitions + old_body
    return ast_transformer.visit(cls_node)


def _generate_field_ast(
    field: ModelField | ModelFieldLike,
    annotation_transformer: "_AnnotationTransformerWithSchemaRenaming",
):
    return ast.Call(
        func=ast.Name("Field"),
        args=[],
        keywords=[
            ast.keyword(
                arg=attr,
                value=ast.parse(
                    annotation_transformer.visit(_get_attribute_from_field_info(field, attr)),
                    mode="eval",
                ).body,
            )
            for attr in _get_passed_attributes_to_field(field.field_info)
        ],
    )


def _get_attribute_from_field_info(field: ModelField | ModelFieldLike, attr: str) -> Any:
    field_value = getattr(field.field_info, attr, Sentinel)
    if field_value is Sentinel:
        field_value = field.field_info.extra.get(attr, Sentinel)
    if field_value is Sentinel:  # pragma: no cover # This is just a safeguard that will most likely never be triggered
        raise CodeGenerationError(f'Field "{attr}" is not present in "{field.name}"')
    return field_value


def _modify_enum_cls(cls_node: ast.ClassDef, enum_info: dict[str, Any]) -> ast.ClassDef:
    transformer = _AnnotationTransformer()
    new_body = [
        ast.Assign(
            targets=[ast.Name(member, ctx=ast.Store())],
            value=ast.Name(transformer.visit(member_value)),
            lineno=0,
        )
        for member, member_value in enum_info.items()
    ]

    old_body = [n for n in cls_node.body if not isinstance(n, ast.AnnAssign | ast.Assign | ast.Pass | ast.Ellipsis)]
    docstring = _pop_docstring_from_cls_body(old_body)

    cls_node.body = docstring + new_body + old_body
    return cls_node


def _pop_docstring_from_cls_body(old_body: list[ast.stmt]) -> list[ast.stmt]:
    if (
        len(old_body) > 0
        and isinstance(old_body[0], ast.Expr)
        and isinstance(old_body[0].value, ast.Constant)
        and isinstance(old_body[0].value.value, str)
    ):
        return [old_body.pop(0)]
    else:
        return []


def _get_passed_attributes_to_field(field_info: FieldInfo):
    for attr_name, attr_val in _dict_of_empty_field_info.items():
        if attr_name == "extra":
            continue
        if getattr(field_info, attr_name) != attr_val:
            yield attr_name
    yield from field_info.extra


class _AnnotationASTNodeTransformer(ast.NodeTransformer):
    def __init__(
        self,
        modified_schemas: dict[str, ModelInfo],
        all_names_in_module: dict[str, str],
        module_python_path: str,
    ):
        super().__init__()
        self.modified_schemas = modified_schemas
        self.module_python_path = module_python_path
        self.all_names_in_module = all_names_in_module

    def visit_Name(self, node: ast.Name) -> Any:  # noqa: N802
        return self._get_name(node, node.id)

    def _get_name(self, node: ast.AST, name: str):
        model_info = self.modified_schemas.get(f"{self.all_names_in_module.get(name, self.module_python_path)}.{name}")
        if model_info is not None:
            return ast.Name(model_info.name)
        return node


class _AnnotationTransformer:
    def visit(self, value: Any):
        if isinstance(value, list | tuple | set | frozenset):
            return self.transform_collection(value)
        if isinstance(value, dict):
            return self.transform_dict(value)
        if isinstance(value, _BaseGenericAlias | GenericAlias):
            return self.transform_generic_alias(value)
        if value is None or value is NoneType:
            return self.transform_none(value)
        if isinstance(value, type):
            return self.transform_type(value)
        if isinstance(value, Enum):
            return self.transform_enum(value)
        if isinstance(value, auto):
            return self.transform_auto(value)
        if isinstance(value, UnionType):
            return self.transform_union(value)
        if isinstance(value, LambdaType) and _LambdaFunctionName == value.__name__:
            return self.transform_lambda(value)
        if inspect.isfunction(value):
            return self.transform_function(value)
        else:
            return self.transform_other(value)

    def transform_collection(self, value: list | tuple | set | frozenset) -> Any:
        return PlainRepr(value.__class__(map(self.visit, value)))

    def transform_dict(self, value: dict) -> Any:
        return PlainRepr(
            value.__class__((self.visit(k), self.visit(v)) for k, v in value.items()),
        )

    def transform_generic_alias(self, value: _BaseGenericAlias | GenericAlias) -> Any:
        return f"{self.visit(get_origin(value))}[{', '.join(self.visit(a) for a in get_args(value))}]"

    def transform_none(self, value: NoneType) -> Any:
        return "None"

    def transform_type(self, value: type) -> Any:
        # NOTE: Be wary of this hack when migrating to pydantic v2
        # This is a hack for pydantic's Constrained types
        if value.__name__.startswith("Constrained") and value.__name__.endswith("Value"):
            # No, get_origin and get_args don't work here. No idea why
            snake_case = _RE_CAMEL_TO_SNAKE.sub("_", value.__name__)
            cls_name = "con" + "".join(snake_case.split("_")[1:-1])
            return (
                cls_name.lower()
                + "("
                + ", ".join(
                    [
                        f"{key}={self.visit(val)}"
                        for key, val in value.__dict__.items()
                        if not key.startswith("__") and val is not None
                    ],
                )
                + ")"
            )
        return value.__name__

    def transform_enum(self, value: Enum) -> Any:
        return PlainRepr(f"{value.__class__.__name__}.{value.name}")

    def transform_auto(self, value: auto) -> Any:
        return PlainRepr("auto()")

    def transform_union(self, value: UnionType) -> Any:
        return "typing.Union[" + (", ".join(self.visit(a) for a in get_args(value))) + "]"

    def transform_lambda(self, value: LambdaType) -> Any:
        # We clean source because getsource() can return only a part of the expression which
        # on its own is not a valid expression such as: "\n  .had(default_factory=lambda: 91)"
        return _find_a_lambda(inspect.getsource(value).strip(" \n\t."))

    def transform_function(self, value: Callable) -> Any:
        return PlainRepr(value.__name__)

    def transform_other(self, value: Any) -> Any:
        return PlainRepr(repr(value))


class PlainRepr(str):
    """
    String class where repr doesn't include quotes.
    """

    def __repr__(self) -> str:
        return str(self)


class _AnnotationTransformerWithSchemaRenaming(_AnnotationTransformer):
    def __init__(
        self,
        modified_schemas: dict[str, ModelInfo],
        module_python_path: str,
        all_names_in_module: dict[str, str],
    ):
        super().__init__()

        self.modified_schemas = modified_schemas
        self.module_python_path = module_python_path
        self.all_names_in_module = all_names_in_module

    def transform_type(self, value: type) -> Any:
        model_info = self.modified_schemas.get(
            f"{self.all_names_in_module.get(value.__name__, self.module_python_path)}.{value.__name__}",
        )
        if model_info is not None:
            return model_info.name
        else:
            return super().transform_type(value)


def _find_a_lambda(source: str) -> str:
    found_lambdas: list[ast.Lambda] = []

    ast.parse(source)
    for node in ast.walk(ast.parse(source)):
        if isinstance(node, ast.keyword) and node.arg == "default_factory" and isinstance(node.value, ast.Lambda):
            found_lambdas.append(node.value)
    if len(found_lambdas) == 1:
        return ast.unparse(found_lambdas[0])
    # These two errors are really hard to cover. Not sure if even possible, honestly :)
    elif len(found_lambdas) == 0:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            f"No lambda found in default_factory even though one was passed: {source}",
        )
    else:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            "More than one lambda found in default_factory. This is not supported.",
        )


# Some day we will want to use this to auto-add imports for new symbols in versions. Some day...
def _get_all_names_defined_in_module(body: ast.Module, module_python_path: str) -> dict[str, str]:
    defined_names = {}
    for node in body.body:
        if isinstance(node, ast.ClassDef | ast.FunctionDef | ast.AsyncFunctionDef):
            defined_names[node.name] = module_python_path
        elif isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    defined_names[target.id] = module_python_path
        elif isinstance(node, ast.ImportFrom):
            for name in node.names:
                defined_names[name.name] = _get_absolute_python_path_of_import(node, module_python_path)
    return defined_names


def _get_cls_pythonpath(cls: type):
    return f"{cls.__module__}.{cls.__name__}"



import datetime
import types
from collections.abc import Mapping
from contextvars import ContextVar
from typing import Any

from fastapi import Depends, Header


def get_cadwyn_dependency(
    *,
    version_header_name: str,
    default_version: datetime.date | None = None,
    extra_kwargs_to_header_constructor: Mapping[str, Any] = types.MappingProxyType({}),
    api_version_var: ContextVar[datetime.date | None] | ContextVar[datetime.date],
) -> Any:
    if default_version is None:
        extra_kwargs: Mapping[str, Any] = extra_kwargs_to_header_constructor
    else:
        extra_kwargs = extra_kwargs_to_header_constructor | {"default": default_version}

    async def dependency(
        api_version: datetime.date = Header(alias=version_header_name, **extra_kwargs),  # noqa: B008
    ):
        api_version_var.set(api_version)
        return api_version

    return Depends(dependency)



import functools
import inspect
import sys
import typing
import warnings
from collections import defaultdict
from collections.abc import Callable, Sequence
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from types import GenericAlias, MappingProxyType, ModuleType
from typing import (
    Any,
    TypeAlias,
    TypeVar,
    _BaseGenericAlias,  # pyright: ignore[reportGeneralTypeIssues]
    cast,
    final,
    get_args,
    get_origin,
)

import fastapi.routing
import fastapi.utils
from fastapi._compat import create_body_model
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import (
    get_body_field,
    get_dependant,
    get_parameterless_sub_dependant,
)
from fastapi.params import Depends
from fastapi.routing import APIRoute
from pydantic import BaseModel
from pydantic.fields import ModelField
from starlette._utils import is_async_callable
from starlette.routing import (
    BaseRoute,
    request_response,
)
from typing_extensions import assert_never

from cadwyn._utils import Sentinel, UnionType, get_another_version_of_module
from cadwyn.codegen import _get_package_path_from_module, _get_version_dir_path
from cadwyn.exceptions import CadwynError, ModuleIsNotVersionedError, RouteAlreadyExistsError, RouterGenerationError
from cadwyn.structure import Version, VersionBundle
from cadwyn.structure.common import Endpoint, VersionDate
from cadwyn.structure.data import _SCHEMA_TO_INTERNAL_REQUEST_BODY_REPRESENTATION_MAPPING
from cadwyn.structure.endpoints import (
    EndpointDidntExistInstruction,
    EndpointExistedInstruction,
    EndpointHadInstruction,
)
from cadwyn.structure.versions import _CADWYN_REQUEST_PARAM_NAME, _CADWYN_RESPONSE_PARAM_NAME, VersionChange

_T = TypeVar("_T", bound=Callable[..., Any])
_R = TypeVar("_R", bound=fastapi.routing.APIRouter)
# This is a hack we do because we can't guarantee how the user will use the router.
_DELETED_ROUTE_TAG = "_CADWYN_DELETED_ROUTE"
EndpointPath: TypeAlias = str
EndpointMethod: TypeAlias = str


@dataclass(slots=True, frozen=True, eq=True)
class _EndpointInfo:
    endpoint_path: str
    endpoint_methods: frozenset[str]


@dataclass(slots=True)
class _RouterInfo:
    router: fastapi.routing.APIRouter
    routes_with_migrated_requests: dict[EndpointPath, set[EndpointMethod]]
    route_bodies_with_migrated_requests: set[type[BaseModel]]


def generate_versioned_routers(
    router: _R,
    versions: VersionBundle,
    latest_schemas_module: ModuleType,
) -> dict[VersionDate, _R]:
    return _EndpointTransformer(router, versions, latest_schemas_module).transform()


class VersionedAPIRouter(fastapi.routing.APIRouter):
    def only_exists_in_older_versions(self, endpoint: _T) -> _T:
        route = _get_route_from_func(self.routes, endpoint)
        if route is None:
            raise LookupError(
                f'Route not found on endpoint: "{endpoint.__name__}". '
                "Are you sure it's a route and decorators are in the correct order?",
            )
        if _DELETED_ROUTE_TAG in route.tags:
            raise CadwynError(f'The route "{endpoint.__name__}" was already deleted. You can\'t delete it again.')
        route.tags.append(_DELETED_ROUTE_TAG)
        return endpoint


@final
class _EndpointTransformer:
    def __init__(
        self,
        parent_router: fastapi.routing.APIRouter,
        versions: VersionBundle,
        latest_schemas_module: ModuleType,
    ) -> None:
        self.parent_router = parent_router
        self.versions = versions
        self.annotation_transformer = _AnnotationTransformer(latest_schemas_module, versions)

        self.routes_that_never_existed = [
            route for route in parent_router.routes if isinstance(route, APIRoute) and _DELETED_ROUTE_TAG in route.tags
        ]

    def transform(self):
        router = deepcopy(self.parent_router)
        router_infos: dict[VersionDate, _RouterInfo] = {}
        routes_with_migrated_requests = {}
        route_bodies_with_migrated_requests: set[type[BaseModel]] = set()
        for version in self.versions:
            #
            self.annotation_transformer.migrate_router_to_version(router, version)

            router_infos[version.value] = _RouterInfo(
                router,
                routes_with_migrated_requests,
                route_bodies_with_migrated_requests,
            )
            # Applying changes for the next version
            routes_with_migrated_requests = _get_migrated_routes_by_path(version)
            route_bodies_with_migrated_requests = {
                schema for change in version.version_changes for schema in change.alter_request_by_schema_instructions
            }
            router = deepcopy(router)
            self._apply_endpoint_changes_to_router(router, version)

        if self.routes_that_never_existed:
            raise RouterGenerationError(
                "Every route you mark with "
                f"@VersionedAPIRouter.{VersionedAPIRouter.only_exists_in_older_versions.__name__} "
                "must be restored in one of the older versions. Otherwise you just need to delete it altogether. "
                "The following routes have been marked with that decorator but were never restored: "
                f"{self.routes_that_never_existed}",
            )

        # BEWARE: We assume that the order of routes didn't change.
        # TODO: Make a test suite checking that it doesn't change
        for route_index, latest_route in enumerate(self.parent_router.routes):
            if not isinstance(latest_route, APIRoute):
                continue
            _add_request_and_response_params(latest_route)
            copy_of_dependant = deepcopy(latest_route.dependant)
            # Remember this: if len(body_params) == 1, then route.body_schema == route.dependant.body_params[0]
            if len(copy_of_dependant.body_params) == 1:
                body_param: ModelField = cast(ModelField, copy_of_dependant.body_params[0])
                body_schema = body_param.type_
                # TODO: Verify that this doesn't break at pydantic 2
                new_type = _SCHEMA_TO_INTERNAL_REQUEST_BODY_REPRESENTATION_MAPPING.get(body_schema, body_schema)
                new_body_param = ModelField(
                    name=body_param.name,
                    type_=new_type,
                    class_validators=body_param.class_validators,
                    model_config=body_param.model_config,
                    default=body_param.default,
                    default_factory=body_param.default_factory,
                    required=body_param.required,
                    final=body_param.final,
                    alias=body_param.alias if body_param.has_alias else None,
                    field_info=body_param.field_info,
                )
                copy_of_dependant.body_params = [new_body_param]  # pyright: ignore[reportGeneralTypeIssues]

            for older_router_info in list(router_infos.values()):
                older_route = older_router_info.router.routes[route_index]

                # We know they are APIRoutes because of the check at the very beginning of the top loop.
                # I.e. Because latest_route is an APIRoute, both routes are  APIRoutes too
                older_route = cast(APIRoute, older_route)
                # Wait.. Why do we need this code again?
                if older_route.body_field is not None and len(older_route.dependant.body_params) == 1:
                    template_older_body_model = self.annotation_transformer._change_version_of_annotations(
                        older_route.body_field.type_,
                        self.annotation_transformer.template_version_dir,
                    )
                else:
                    template_older_body_model = None
                _add_data_migrations_to_route(
                    older_route,
                    template_older_body_model,
                    older_route.body_field.alias if older_route.body_field is not None else None,
                    copy_of_dependant,
                    # NOTE: The fact that we use latest here assumes that the route can never change its response schema
                    latest_route.response_model,
                    self.versions,
                )
        for _, router_info in router_infos.items():
            router_info.router.routes = [
                route
                for route in router_info.router.routes
                if not (isinstance(route, fastapi.routing.APIRoute) and _DELETED_ROUTE_TAG in route.tags)
            ]
        return {version: router_info.router for version, router_info in router_infos.items()}

    # TODO: Simplify https://github.com/zmievsa/cadwyn/issues/28
    def _apply_endpoint_changes_to_router(  # noqa: C901
        self,
        router: fastapi.routing.APIRouter,
        version: Version,
    ):
        routes = router.routes
        for version_change in version.version_changes:
            for instruction in version_change.alter_endpoint_instructions:
                original_routes = _get_routes(
                    routes,
                    instruction.endpoint_path,
                    instruction.endpoint_methods,
                    instruction.endpoint_func_name,
                    is_deleted=False,
                )
                methods_to_which_we_applied_changes = set()
                methods_we_should_have_applied_changes_to = instruction.endpoint_methods.copy()

                if isinstance(instruction, EndpointDidntExistInstruction):
                    # TODO OPTIMIZATION:
                    deleted_routes = _get_routes(
                        routes,
                        instruction.endpoint_path,
                        instruction.endpoint_methods,
                        instruction.endpoint_func_name,
                        is_deleted=True,
                    )
                    if deleted_routes:
                        method_union = set()
                        for deleted_route in deleted_routes:
                            method_union |= deleted_route.methods
                        raise RouterGenerationError(
                            f'Endpoint "{list(method_union)} {instruction.endpoint_path}" you tried to delete in '
                            f'"{version_change.__name__}" was already deleted in a newer version. If you really have '
                            f'two routes with the same paths and methods, please, use "endpoint(..., func_name=...)" '
                            f"to distinguish between them. Function names of endpoints that were already deleted: "
                            f"{[r.endpoint.__name__ for r in deleted_routes]}",
                        )
                    for original_route in original_routes:
                        methods_to_which_we_applied_changes |= original_route.methods
                        original_route.tags.append(_DELETED_ROUTE_TAG)
                    err = (
                        'Endpoint "{endpoint_methods} {endpoint_path}" you tried to delete in'
                        ' "{version_change_name}" doesn\'t exist in a newer version'
                    )
                elif isinstance(instruction, EndpointExistedInstruction):
                    # TODO Optimization
                    if original_routes:
                        method_union = set()
                        for original_route in original_routes:
                            method_union |= original_route.methods
                        raise RouterGenerationError(
                            f'Endpoint "{list(method_union)} {instruction.endpoint_path}" you tried to restore in'
                            f' "{version_change.__name__}" already existed in a newer version. If you really have two '
                            f'routes with the same paths and methods, please, use "endpoint(..., func_name=...)" to '
                            f"distinguish between them. Function names of endpoints that already existed: "
                            f"{[r.endpoint.__name__ for r in original_routes]}",
                        )
                    deleted_routes = _get_routes(
                        routes,
                        instruction.endpoint_path,
                        instruction.endpoint_methods,
                        instruction.endpoint_func_name,
                        is_deleted=True,
                    )
                    try:
                        _validate_no_repetitions_in_routes(deleted_routes)
                    except RouteAlreadyExistsError as e:
                        raise RouterGenerationError(
                            f'Endpoint "{list(instruction.endpoint_methods)} {instruction.endpoint_path}" you tried to '
                            f'restore in "{version_change.__name__}" has {len(e.routes)} applicable routes that could '
                            f"be restored. If you really have two routes with the same paths and methods, please, use "
                            f'"endpoint(..., func_name=...)" to distinguish between them. Function names of '
                            f"endpoints that can be restored: {[r.endpoint.__name__ for r in e.routes]}",
                        ) from e
                    for deleted_route in deleted_routes:
                        methods_to_which_we_applied_changes |= deleted_route.methods
                        deleted_route.tags.remove(_DELETED_ROUTE_TAG)

                        routes_that_never_existed = _get_routes(
                            self.routes_that_never_existed,
                            deleted_route.path,
                            deleted_route.methods,
                            deleted_route.endpoint.__name__,
                            is_deleted=True,
                        )
                        if len(routes_that_never_existed) == 1:
                            self.routes_that_never_existed.remove(routes_that_never_existed[0])
                        elif len(routes_that_never_existed) > 1:  # pragma: no cover
                            # I am not sure if it's possible to get to this error but I also don't want
                            # to remove it because I like its clarity very much
                            routes = routes_that_never_existed
                            raise RouterGenerationError(
                                f'Endpoint "{list(deleted_route.methods)} {deleted_route.path}" you tried to restore '
                                f'in "{version_change.__name__}" has {len(routes_that_never_existed)} applicable '
                                f"routes with the same function name and path that could be restored. This can cause "
                                f"problems during version generation. Specifically, Cadwyn won't be able to warn "
                                f"you when you deleted a route and never restored it. Please, make sure that "
                                f"functions for all these routes have different names: "
                                f"{[f'{r.endpoint.__module__}.{r.endpoint.__name__}' for r in routes]}",
                            )
                    err = (
                        'Endpoint "{endpoint_methods} {endpoint_path}" you tried to restore in'
                        ' "{version_change_name}" wasn\'t among the deleted routes'
                    )
                elif isinstance(instruction, EndpointHadInstruction):
                    for original_route in original_routes:
                        methods_to_which_we_applied_changes |= original_route.methods
                        _apply_endpoint_had_instruction(version_change, instruction, original_route)
                    err = (
                        'Endpoint "{endpoint_methods} {endpoint_path}" you tried to change in'
                        ' "{version_change_name}" doesn\'t exist'
                    )
                else:
                    assert_never(instruction)
                method_diff = methods_we_should_have_applied_changes_to - methods_to_which_we_applied_changes
                if method_diff:
                    raise RouterGenerationError(
                        err.format(
                            endpoint_methods=list(method_diff),
                            endpoint_path=instruction.endpoint_path,
                            version_change_name=version_change.__name__,
                        ),
                    )


def _validate_no_repetitions_in_routes(routes: list[fastapi.routing.APIRoute]):
    route_map = {}

    for route in routes:
        route_info = _EndpointInfo(route.path, frozenset(route.methods))
        if route_info in route_map:
            raise RouteAlreadyExistsError(route, route_map[route_info])
        route_map[route_info] = route


@final
class _AnnotationTransformer:
    __slots__ = (
        "latest_schemas_module",
        "version_dirs",
        "template_version_dir",
        "latest_version_dir",
        "change_versions_of_a_non_container_annotation",
    )

    def __init__(self, latest_schemas_module: ModuleType, versions: VersionBundle) -> None:
        if not hasattr(latest_schemas_module, "__path__"):
            raise RouterGenerationError(
                f'The latest schemas module must be a package. "{latest_schemas_module.__name__}" is not a package.',
            )
        if not latest_schemas_module.__name__.endswith(".latest"):
            raise RouterGenerationError(
                'The name of the latest schemas module must be "latest". '
                f'Received "{latest_schemas_module.__name__}" instead.',
            )
        self.latest_schemas_module = latest_schemas_module
        self.version_dirs = frozenset(
            [_get_package_path_from_module(latest_schemas_module)]
            + [_get_version_dir_path(latest_schemas_module, version.value) for version in versions],
        )
        # Okay, the naming is confusing, I know. Essentially template_version_dir is a dir of
        # latest_schemas_module while latest_version_dir is a version equivalent to latest but
        # with its own directory. Pick a better naming and make a PR, I am at your mercy.
        self.template_version_dir = min(self.version_dirs)  # "latest" < "v0000_00_00"
        self.latest_version_dir = max(self.version_dirs)  # "v2005_11_11" > "v2000_11_11"

        # This cache is not here for speeding things up. It's for preventing the creation of copies of the same object
        # because such copies could produce weird behaviors at runtime, especially if you/fastapi do any comparisons.
        # It's defined here and not on the method because of this: https://youtu.be/sVjtp6tGo0g
        self.change_versions_of_a_non_container_annotation = functools.cache(
            self._change_versions_of_a_non_container_annotation,
        )

    def migrate_router_to_version(self, router: fastapi.routing.APIRouter, version: Version):
        version_dir = _get_version_dir_path(self.latest_schemas_module, version.value)
        if not version_dir.is_dir():
            raise RouterGenerationError(
                f"Versioned schema directory '{version_dir}' does not exist.",
            )
        for route in router.routes:
            if not isinstance(route, fastapi.routing.APIRoute):
                continue
            self.migrate_route_to_version(route, version_dir)

    def migrate_route_to_version(
        self,
        route: fastapi.routing.APIRoute,
        version_dir: Path,
        *,
        ignore_response_model: bool = False,
    ):
        if route.response_model is not None and not ignore_response_model:
            route.response_model = self._change_version_of_annotations(route.response_model, version_dir)
            route.response_field = fastapi.utils.create_response_field(
                name="Response_" + route.unique_id,
                type_=route.response_model,
                mode="serialization",
            )
            route.secure_cloned_response_field = fastapi.utils.create_cloned_field(route.response_field)
        route.dependencies = self._change_version_of_annotations(route.dependencies, version_dir)
        route.endpoint = self._change_version_of_annotations(route.endpoint, version_dir)
        for callback in route.callbacks or []:
            if not isinstance(callback, APIRoute):
                continue
            self.migrate_route_to_version(callback, version_dir, ignore_response_model=ignore_response_model)
        _remake_endpoint_dependencies(route)

    def get_another_version_of_cls(self, cls_from_old_version: type[Any], new_version_dir: Path):
        # version_dir = /home/myuser/package/companies/v2021_01_01
        module_from_old_version = sys.modules[cls_from_old_version.__module__]
        try:
            module = get_another_version_of_module(module_from_old_version, new_version_dir, self.version_dirs)
        except ModuleIsNotVersionedError:
            return cls_from_old_version
        return getattr(module, cls_from_old_version.__name__)

    def _change_versions_of_a_non_container_annotation(self, annotation: Any, version_dir: Path) -> Any:
        if isinstance(annotation, _BaseGenericAlias | GenericAlias):
            return get_origin(annotation)[
                tuple(self._change_version_of_annotations(arg, version_dir) for arg in get_args(annotation))
            ]
        elif isinstance(annotation, Depends):
            return Depends(
                self._change_version_of_annotations(annotation.dependency, version_dir),
                use_cache=annotation.use_cache,
            )
        elif isinstance(annotation, UnionType):
            getitem = typing.Union.__getitem__  # pyright: ignore[reportGeneralTypeIssues]
            return getitem(
                tuple(self._change_version_of_annotations(a, version_dir) for a in get_args(annotation)),
            )
        elif annotation is typing.Any or isinstance(annotation, typing.NewType):
            return annotation
        elif isinstance(annotation, type):
            if annotation.__module__ == "pydantic.main" and issubclass(annotation, BaseModel):
                return create_body_model(
                    fields=self._change_version_of_annotations(annotation.__fields__, version_dir).values(),
                    model_name=annotation.__name__,
                )
            return self._change_version_of_type(annotation, version_dir)
        elif callable(annotation):
            # TASK: https://github.com/zmievsa/cadwyn/issues/48
            if inspect.iscoroutinefunction(annotation):

                @functools.wraps(annotation)
                async def new_callable(  # pyright: ignore[reportGeneralTypeIssues]
                    *args: Any,
                    **kwargs: Any,
                ) -> Any:
                    return await annotation(*args, **kwargs)

            else:

                @functools.wraps(annotation)
                def new_callable(  # pyright: ignore[reportGeneralTypeIssues]
                    *args: Any,
                    **kwargs: Any,
                ) -> Any:
                    return annotation(*args, **kwargs)

            # Otherwise it will have the same signature as __wrapped__
            new_callable.__alt_wrapped__ = new_callable.__wrapped__  # pyright: ignore[reportGeneralTypeIssues]
            del new_callable.__wrapped__
            old_params = inspect.signature(annotation).parameters
            callable_annotations = new_callable.__annotations__

            new_callable: Any = cast(Any, new_callable)
            new_callable.__annotations__ = self._change_version_of_annotations(
                callable_annotations,
                version_dir,
            )
            new_callable.__defaults__ = self._change_version_of_annotations(
                tuple(p.default for p in old_params.values() if p.default is not inspect.Signature.empty),
                version_dir,
            )
            new_callable.__signature__ = _generate_signature(new_callable, old_params)
            return new_callable
        else:
            return annotation

    def _change_version_of_annotations(self, annotation: Any, version_dir: Path) -> Any:
        """Recursively go through all annotations and if they were taken from any versioned package, change them to the
        annotations corresponding to the version_dir passed.

        So if we had a annotation "UserResponse" from "latest" version, and we passed version_dir of "v1_0_1", it would
        replace "UserResponse" with the the same class but from the "v1_0_1" version.

        """
        if isinstance(annotation, dict):
            return {
                self._change_version_of_annotations(key, version_dir): self._change_version_of_annotations(
                    value,
                    version_dir,
                )
                for key, value in annotation.items()
            }

        elif isinstance(annotation, list | tuple):
            return type(annotation)(self._change_version_of_annotations(v, version_dir) for v in annotation)
        else:
            return self.change_versions_of_a_non_container_annotation(annotation, version_dir)

    def _change_version_of_type(self, annotation: type, version_dir: Path):
        if issubclass(annotation, BaseModel | Enum):
            if version_dir == self.latest_version_dir:
                source_file = inspect.getsourcefile(annotation)
                if source_file is None:  # pragma: no cover # I am not even sure how to cover this
                    warnings.warn(
                        f'Failed to find where the type annotation "{annotation}" is located.'
                        "Please, double check that it's located in the right directory",
                        stacklevel=7,
                    )
                else:
                    self._validate_source_file_is_located_in_template_dir(annotation, source_file)
            return self.get_another_version_of_cls(annotation, version_dir)
        else:
            return annotation

    def _validate_source_file_is_located_in_template_dir(self, annotation: type, source_file: str):
        template_dir = str(self.template_version_dir)
        dir_with_versions = str(self.template_version_dir.parent)
        # So if it is somewhere close to version dirs (either within them or next to them),
        # but not located in "latest",
        # but also not located in any other version dir
        if (
            source_file.startswith(dir_with_versions)
            and not source_file.startswith(template_dir)
            and any(source_file.startswith(str(d)) for d in self.version_dirs)
        ):
            raise RouterGenerationError(
                f'"{annotation}" is not defined in "{self.template_version_dir}" even though it must be. '
                f'It is defined in "{Path(source_file).parent}". '
                "It probably means that you used a specific version of the class in fastapi dependencies "
                'or pydantic schemas instead of "latest".',
            )


def _remake_endpoint_dependencies(route: fastapi.routing.APIRoute):
    route.dependant = get_dependant(path=route.path_format, call=route.endpoint)
    _add_request_and_response_params(route)
    route.body_field = get_body_field(dependant=route.dependant, name=route.unique_id)
    for depends in route.dependencies[::-1]:
        route.dependant.dependencies.insert(
            0,
            get_parameterless_sub_dependant(depends=depends, path=route.path_format),
        )
    route.app = request_response(route.get_route_handler())


def _add_request_and_response_params(route: APIRoute):
    if not route.dependant.request_param_name:
        route.dependant.request_param_name = _CADWYN_REQUEST_PARAM_NAME
    if not route.dependant.response_param_name:
        route.dependant.response_param_name = _CADWYN_RESPONSE_PARAM_NAME


def _add_data_migrations_to_route(
    route: APIRoute,
    template_body_field: type[BaseModel] | None,
    template_body_field_name: str | None,
    dependant_for_request_migrations: Dependant,
    latest_response_model: Any,
    versions: VersionBundle,
):
    if not is_async_callable(route.endpoint):
        raise RouterGenerationError(
            f'All versioned endpoints must be asynchronous. Endpoint "{route.endpoint}" is not.',
        )
    if not (route.dependant.request_param_name and route.dependant.response_param_name):  # pragma: no cover
        raise CadwynError(
            f"{route.dependant.request_param_name=}, {route.dependant.response_param_name=} "
            f"for route {list(route.methods)} {route.path} which should not be possible. Please, contact my author.",
        )

    route.endpoint = versions._versioned(
        template_body_field,
        template_body_field_name,
        route.dependant.body_params,
        dependant_for_request_migrations,
        latest_response_model,
        request_param_name=route.dependant.request_param_name,
        response_param_name=route.dependant.response_param_name,
    )(route.endpoint)
    route.dependant.call = route.endpoint


def _apply_endpoint_had_instruction(
    version_change: type[VersionChange],
    instruction: EndpointHadInstruction,
    original_route: APIRoute,
):
    for attr_name in instruction.attributes.__dataclass_fields__:
        attr = getattr(instruction.attributes, attr_name)
        if attr is not Sentinel:
            if getattr(original_route, attr_name) == attr:
                raise RouterGenerationError(
                    f'Expected attribute "{attr_name}" of endpoint'
                    f' "{list(original_route.methods)} {original_route.path}"'
                    f' to be different in "{version_change.__name__}", but it was the same.'
                    " It means that your version change has no effect on the attribute"
                    " and can be removed.",
                )
            setattr(original_route, attr_name, attr)


def _generate_signature(
    new_callable: Callable,
    old_params: MappingProxyType[str, inspect.Parameter],
):
    parameters = []
    default_counter = 0
    for param in old_params.values():
        if param.default is not inspect.Signature.empty:
            default = new_callable.__defaults__[default_counter]
            default_counter += 1
        else:
            default = inspect.Signature.empty
        parameters.append(
            inspect.Parameter(
                param.name,
                param.kind,
                default=default,
                annotation=new_callable.__annotations__.get(
                    param.name,
                    inspect.Signature.empty,
                ),
            ),
        )
    return inspect.Signature(
        parameters=parameters,
        return_annotation=new_callable.__annotations__.get(
            "return",
            inspect.Signature.empty,
        ),
    )


def _get_routes(
    routes: Sequence[BaseRoute],
    endpoint_path: str,
    endpoint_methods: set[str],
    endpoint_func_name: str | None = None,
    *,
    is_deleted: bool = False,
) -> list[fastapi.routing.APIRoute]:
    found_routes = []
    for route in routes:
        if (
            isinstance(route, fastapi.routing.APIRoute)
            and route.path == endpoint_path
            and set(route.methods).issubset(endpoint_methods)
            and (endpoint_func_name is None or route.endpoint.__name__ == endpoint_func_name)
            and (_DELETED_ROUTE_TAG in route.tags) == is_deleted
        ):
            found_routes.append(route)
    return found_routes


def _get_route_from_func(
    routes: Sequence[BaseRoute],
    endpoint: Endpoint,
) -> fastapi.routing.APIRoute | None:
    for route in routes:
        if isinstance(route, fastapi.routing.APIRoute) and (route.endpoint == endpoint):
            return route
    return None


def _get_migrated_routes_by_path(version: Version) -> dict[EndpointPath, set[EndpointMethod]]:
    request_by_path_migration_instructions = [
        version_change.alter_request_by_path_instructions for version_change in version.version_changes
    ]
    migrated_routes = defaultdict(set)
    for instruction_dict in request_by_path_migration_instructions:
        for path, instruction_list in instruction_dict.items():
            for instruction in instruction_list:
                migrated_routes[path] |= instruction.methods
    return migrated_routes



from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo

from cadwyn.exceptions import CadwynStructureError

from .._utils import Sentinel

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny


@dataclass(slots=True)
class FieldChanges:
    default: Any
    default_factory: Any
    alias: str
    title: str
    description: str
    exclude: "AbstractSetIntStr | MappingIntStrAny | Any"
    include: "AbstractSetIntStr | MappingIntStrAny | Any"
    const: bool
    gt: float
    ge: float
    lt: float
    le: float
    multiple_of: float
    allow_inf_nan: bool
    max_digits: int
    decimal_places: int
    min_items: int
    max_items: int
    unique_items: bool
    min_length: int
    max_length: int
    allow_mutation: bool
    regex: str
    discriminator: str
    repr: bool


@dataclass(slots=True)
class OldSchemaFieldHad:
    schema: type[BaseModel]
    field_name: str
    type: type
    field_changes: FieldChanges
    new_name: str


@dataclass(slots=True)
class OldSchemaFieldDidntExist:
    schema: type[BaseModel]
    field_name: str


@dataclass(slots=True)
class OldSchemaFieldExistedWith:
    schema: type[BaseModel]
    field_name: str
    type: type
    field: FieldInfo
    import_from: str | None = None
    import_as: str | None = None

    def __post_init__(self):
        if self.import_from is None and self.import_as is not None:
            raise CadwynStructureError(
                f'Field "{self.field_name}" has "import_as" but not "import_from" which is prohibited',
            )


@dataclass(slots=True)
class AlterFieldInstructionFactory:
    schema: type[BaseModel]
    name: str

    def had(
        self,
        *,
        name: str = Sentinel,
        type: Any = Sentinel,
        default: Any = Sentinel,
        default_factory: Callable = Sentinel,
        alias: str = Sentinel,
        title: str = Sentinel,
        description: str = Sentinel,
        exclude: "AbstractSetIntStr | MappingIntStrAny | Any" = Sentinel,
        include: "AbstractSetIntStr | MappingIntStrAny | Any" = Sentinel,
        const: bool = Sentinel,
        gt: float = Sentinel,
        ge: float = Sentinel,
        lt: float = Sentinel,
        le: float = Sentinel,
        multiple_of: float = Sentinel,
        allow_inf_nan: bool = Sentinel,
        max_digits: int = Sentinel,
        decimal_places: int = Sentinel,
        min_items: int = Sentinel,
        max_items: int = Sentinel,
        unique_items: bool = Sentinel,
        min_length: int = Sentinel,
        max_length: int = Sentinel,
        allow_mutation: bool = Sentinel,
        regex: str = Sentinel,
        discriminator: str = Sentinel,
        repr: bool = Sentinel,
    ) -> OldSchemaFieldHad:
        return OldSchemaFieldHad(
            schema=self.schema,
            field_name=self.name,
            type=type,
            new_name=name,
            field_changes=FieldChanges(
                default=default,
                default_factory=default_factory,
                alias=alias,
                title=title,
                description=description,
                exclude=exclude,
                include=include,
                const=const,
                gt=gt,
                ge=ge,
                lt=lt,
                le=le,
                multiple_of=multiple_of,
                allow_inf_nan=allow_inf_nan,
                max_digits=max_digits,
                decimal_places=decimal_places,
                min_items=min_items,
                max_items=max_items,
                unique_items=unique_items,
                min_length=min_length,
                max_length=max_length,
                allow_mutation=allow_mutation,
                regex=regex,
                discriminator=discriminator,
                repr=repr,
            ),
        )

    @property
    def didnt_exist(self) -> OldSchemaFieldDidntExist:
        return OldSchemaFieldDidntExist(self.schema, field_name=self.name)

    def existed_as(
        self,
        *,
        type: Any,
        info: FieldInfo | None = None,
        import_from: str | None = None,
        import_as: str | None = None,
    ) -> OldSchemaFieldExistedWith:
        return OldSchemaFieldExistedWith(
            self.schema,
            field_name=self.name,
            type=type,
            field=info or Field(),
            import_from=import_from,
            import_as=import_as,
        )


AlterSchemaSubInstruction = OldSchemaFieldHad | OldSchemaFieldDidntExist | OldSchemaFieldExistedWith


@dataclass(slots=True)
class AlterSchemaInstruction:
    schema: type[BaseModel]
    name: str


@dataclass(slots=True)
class AlterSchemaSubInstructionFactory:
    schema: type[BaseModel]

    def field(self, name: str, /) -> AlterFieldInstructionFactory:
        return AlterFieldInstructionFactory(self.schema, name)

    def had(self, *, name: str) -> AlterSchemaInstruction:
        return AlterSchemaInstruction(self.schema, name)


def schema(model: type[BaseModel], /) -> AlterSchemaSubInstructionFactory:
    return AlterSchemaSubInstructionFactory(model)



import datetime
from collections.abc import Awaitable, Callable
from typing import ParamSpec, TypeAlias, TypeVar

from pydantic import BaseModel

VersionedModel = BaseModel
VersionDate = datetime.date
_P = ParamSpec("_P")
_R = TypeVar("_R")
Endpoint: TypeAlias = Callable[_P, Awaitable[_R]]



from fastapi.routing import APIRoute


class CadwynError(Exception):
    pass


class LintingError(CadwynError):
    pass


class CodeGenerationError(CadwynError):
    pass


class InvalidGenerationInstructionError(CodeGenerationError):
    pass


class RouterGenerationError(CadwynError):
    pass


class RouteAlreadyExistsError(RouterGenerationError):
    def __init__(self, *routes: APIRoute):
        self.routes = routes
        super().__init__(f"The following routes are duplicates of each other: {routes}")


class CadwynStructureError(CadwynError):
    pass


class ModuleIsNotVersionedError(ValueError):
    pass
